# -*- coding: utf-8 -*-

"""Anaconda Navigator conda interface."""
from __future__ import annotations

import functools
import json
import os
import typing

from navigator_updater.config import NAVIGATOR_LOCKFILE
from navigator_updater.external import filelock
from navigator_updater.utils.conda import exceptions
from navigator_updater.utils.conda import types as conda_types
from navigator_updater.utils.conda.core import get_conda_info
from navigator_updater.utils.conda.launch import run_process, launch_process, ProcessOutput


CondaSearchResult = typing.MutableMapping[str, typing.List[conda_types.CondaSearchPackageInfo]]


class CondaAPI:
    """An interface for interacting with conda."""

    @functools.cached_property
    def conda_info(self) -> conda_types.CondaInfoOutput:
        """Returns Conda information in a dictionary format."""
        info: conda_types.CondaInfoOutput | None = get_conda_info()
        if info is None:
            raise Exception('Conda info is not available!')  # pylint: disable=broad-exception-raised

        return info

    @property
    def root_prefix(self) -> str:
        """Returns the root prefix path where Conda is installed."""
        return self.conda_info.get('root_prefix', '')

    @property
    def conda_prefix(self) -> str:
        """Returns the Conda prefix path."""
        return self.conda_info.get('conda_prefix', self.root_prefix)

    @property
    def active_prefix(self) -> str:
        """Returns the active Conda environment prefix path."""
        return self.conda_info.get('active_prefix') or self.root_prefix

    @property
    def root_writable(self) -> bool:
        """Returns True if the root environment is writable; False otherwise."""
        return self.conda_info.get('root_writable', False)

    @functools.lru_cache
    def search(self, package: str) -> typing.Optional[CondaSearchResult]:
        """Searches for a package in conda repositories."""
        cmd: typing.Tuple[str, ...] = ('conda', 'search', package, '--json')
        process: ProcessOutput = run_process(cmd)

        result: typing.Union[CondaSearchResult, conda_types.CondaErrorOutput]
        try:
            result = json.loads(process.stdout)
        except (TypeError, ValueError):
            return None

        if 'error' not in result:
            return typing.cast(CondaSearchResult, result)

        raise exceptions.CondaError(error=typing.cast(conda_types.CondaErrorOutput, result))

    def install(self, package_spec: str) -> None:
        """Launch `conda install` subprocess."""
        cmd: typing.Tuple[str, ...] = ('conda', 'install', '--yes', '--json', '--force-reinstall', package_spec)

        with filelock.FileLock(NAVIGATOR_LOCKFILE):  # type: ignore
            launch_process(cmd)

    def linked(self, prefix: str | None = None) -> typing.Set[str]:
        """Returns a set of canonical names of linked packages in the specified Conda environment."""
        prefix = self.active_prefix if prefix is None else prefix

        if not os.path.isdir(prefix):
            return set()

        packages: typing.Set[str] = set()
        meta_dir: str = os.path.join(prefix, 'conda-meta')

        if os.path.isdir(meta_dir):
            meta_files: typing.Set[str] = set(fname for fname in os.listdir(meta_dir) if fname.endswith('.json'))
            packages = set(fname[:-5] for fname in meta_files)

        return packages



# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# Copyright 2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""Anaconda Navigator Updater main dialog."""

from __future__ import annotations

import os
import threading
import typing

from qtpy.QtCore import QSize, Qt, QTimer, Signal
from qtpy.QtSvg import QSvgWidget
from qtpy.QtWidgets import QHBoxLayout, QVBoxLayout

from navigator_updater.api.conda_api import CondaAPI
from navigator_updater.config import MAC, WIN, NAVIGATOR_LOCKFILE
from navigator_updater.external import filelock
from navigator_updater.static import images
from navigator_updater.utils import launch as launch_utils
from navigator_updater.utils import version_utils
from navigator_updater.utils.conda import types as conda_types
from navigator_updater.utils.misc import set_windows_appusermodelid
from navigator_updater.utils.styles import load_style_sheet
from navigator_updater.widgets import ButtonNormal, ButtonPrimary, LabelBase, SpacerHorizontal, SpacerVertical
from navigator_updater.widgets.dialogs import DialogBase

CondaSearchResult = typing.MutableMapping[str, typing.List[conda_types.CondaSearchPackageInfo]]


def split_canonical_name(cname: str) -> typing.Tuple[str, ...]:
    """Split a canonical package name into name, version, build."""
    return tuple(cname.rsplit('-', 2))


class MainDialog(DialogBase):  # pylint: disable=too-many-instance-attributes
    """Main dialog for the anaconda navgator updater."""

    # Signals
    sig_application_updated = Signal()
    sig_ready = Signal()

    # Class variables
    PACKAGE = 'anaconda-navigator'  # pylint: disable=invalid-name
    WIDTH = 450  # pylint: disable=invalid-name
    HEIGHT = 200  # pylint: disable=invalid-name

    def __init__(self):  # pylint: disable=too-many-statements
        """Main dialog for the anaconda navgator updater."""
        super().__init__()

        # Variables
        self.api = CondaAPI()

        self.current_version = version_utils.Version(None)
        self.latest_version = version_utils.Version(None)
        self.busy = False

        self.style_sheet = load_style_sheet()
        self.lookup_timer = QTimer()
        self.refresh_timer = QTimer()

        # Widgets
        self.label_icon = QSvgWidget()
        self.label_message = LabelBase('We strongly recommend you to keep Anaconda Navigator up to date.')

        self.label_status = LabelBase('')
        self.button_cancel = ButtonNormal('Dismiss')
        self.button_update = ButtonPrimary('Update now')
        self.button_launch = ButtonPrimary('Launch Navigator')

        # Widgets setup

        self._windows_appusermodelid = set_windows_appusermodelid() if WIN else None

        self.label_message.setAlignment(Qt.AlignLeft | Qt.AlignTop)
        self.label_message.setWordWrap(True)
        self.label_status.setWordWrap(True)

        self.button_update.setAutoDefault(True)
        self.button_update.setFocusPolicy(Qt.NoFocus)
        self.button_update.setDisabled(True)
        self.button_launch.setAutoDefault(True)
        self.button_launch.setFocusPolicy(Qt.NoFocus)
        self.button_cancel.setFocusPolicy(Qt.NoFocus)

        self.refresh_timer.setInterval(3000)

        self.label_icon.load(images.ANACONDA_LOGO)
        self.label_icon.setMaximumSize(QSize(64, 64))
        self.label_icon.setMinimumSize(QSize(64, 64))

        self.setWindowTitle('Anaconda Navigator Updater')
        self.setMinimumSize(self.WIDTH, self.HEIGHT)
        self.setMinimumWidth(self.WIDTH)
        self.setMaximumWidth(self.WIDTH)
        self.setMinimumHeight(self.HEIGHT)

        # Layouts
        layout_status = QHBoxLayout()
        layout_status.addWidget(self.label_status)
        layout_status.addWidget(SpacerHorizontal())

        layout_text = QVBoxLayout()
        layout_text.addWidget(self.label_message)
        layout_text.addStretch()
        layout_text.addWidget(SpacerVertical())
        layout_text.addLayout(layout_status)

        layout_icon = QVBoxLayout()
        layout_icon.addWidget(self.label_icon)
        layout_icon.addStretch()

        layout_top = QHBoxLayout()
        layout_top.addLayout(layout_icon)
        layout_top.addWidget(SpacerHorizontal())
        layout_top.addLayout(layout_text)

        layout_buttons = QHBoxLayout()
        layout_buttons.addStretch()
        layout_buttons.addWidget(self.button_cancel)
        layout_buttons.addWidget(SpacerHorizontal())
        layout_buttons.addWidget(self.button_update)
        layout_buttons.addWidget(self.button_launch)

        layout = QVBoxLayout()
        layout.addLayout(layout_top)
        layout.addWidget(SpacerVertical())
        layout.addWidget(SpacerVertical())
        layout.addStretch()
        layout.addLayout(layout_buttons)

        self.navigator_lookup()
        self.setLayout(layout)

        # Signals
        self.button_update.clicked.connect(self.install_update)
        self.button_cancel.clicked.connect(self.reject)
        self.button_launch.clicked.connect(self.launch)

        self.refresh_timer.timeout.connect(self.refresh)
        self.refresh_timer.start()

    @property
    def up_to_date(self) -> bool:
        """Flag latest Anaconda Navigator is installed."""
        return version_utils.compare(self.current_version, self.latest_version) >= 0

    def __set_messages(self, label_message: str, status_message: str | None = None) -> None:
        self.label_message.setText(label_message)
        if status_message is not None:
            self.label_status.setText(status_message)

    def refresh(self):  # pylint: disable=too-many-branches,too-many-statements
        """Refresh enabled/disabled status of widgets."""
        if self.up_to_date:
            self.lookup_timer.stop()

        current_version = str(self.current_version)
        latest_version = str(self.latest_version) or '-'

        self.button_update.setDisabled(True)
        if self.current_version.version is None:
            current_version = 'Not installed'
            self.button_launch.setDisabled(True)

        message = None
        main_message = (
            f'Current version:    &nbsp;&nbsp;&nbsp;&nbsp;<i>{current_version}</i><br>'
            f'Available version: &nbsp;&nbsp;<b>{latest_version}</b><br>'
        )

        if self.check_running():
            if self.up_to_date:
                message = 'Anaconda Navigator is up to date and running.'
            else:
                message = 'Please close Anaconda Navigator before updating.'

            self.button_launch.setDisabled(True)
            self.__set_messages(main_message, message)
            return

        if self.up_to_date:
            message = 'Anaconda Navigator is up to date.'
            self.busy = False
            self.button_launch.setDisabled(False)
            self.__set_messages(main_message, message)
            return

        if not self.api.root_writable and WIN:
            message = 'Need to run with elevated privileges'
            self.__set_messages(main_message, message)
            return

        self.button_update.setDisabled(self.busy)
        self.__set_messages(main_message, message)

    def navigator_lookup(self) -> None:
        """Check installed packages in case conda was used."""
        package = next((package for package in self.api.linked() if self.PACKAGE in package), None)
        self.current_version = version_utils.Version(split_canonical_name(package)[1] if package else None)

        if self.latest_version.version is not None:
            self.refresh()
            return

        search_result = self.api.search(self.PACKAGE)

        packages: typing.List[conda_types.CondaSearchPackageInfo] = []
        if search_result:
            packages = search_result.get(self.PACKAGE, [])

        if packages:
            self.latest_version = version_utils.as_version(version_utils.sort(
                (package.get('version', '') for package in packages), unique=True, reverse=True)[0])

        self.refresh()

    def install_update(self):
        """Start Anaconda Navigator update thread."""
        self.lookup_timer.setInterval(6000)
        self.lookup_timer.timeout.connect(self.navigator_lookup)
        self.lookup_timer.start()

        name = self.api.conda_info.get('active_prefix_name', os.path.basename(self.api.active_prefix))

        if self.latest_version.version is not None:
            msg = f'Installing package on <b>{name}</b>...'
        else:
            msg = f'Updating package on <b>{name}</b>...'

        threading.Thread(target=self.api.install, args=(self.PACKAGE,)).start()
        self.busy = True

        self.label_status.setText(msg)
        self.button_update.setDisabled(True)
        self.button_launch.setDisabled(True)

    def launch(self):
        """Launch Anaconda Navigator."""
        command = 'anaconda-navigator'

        # Use the app bundle on OSX
        if MAC:
            command = f'open \'{os.path.join(self.prefix, "Anaconda-Navigator.app")}\''

        launch_utils.launch(
            root_prefix=self.api.root_prefix,
            prefix=self.api.conda_prefix,
            command=command,
            package_name='anaconda-navigator-app'
        )
        self.close()

    @staticmethod
    def check_running() -> bool:
        """Check if Anaconda Navigator is running."""
        # Create file lock
        lock = filelock.FileLock(NAVIGATOR_LOCKFILE)  # type: ignore
        running = False
        try:
            with lock.acquire(timeout=0.01):
                pass
        except filelock.Timeout:  # type: ignore
            running = True
        return running



# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# Copyright (c) Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""Miscellaneous utilities."""

import os
import os.path as osp

from navigator_updater.config import PIDFILE


def __remove_pyc_pyo(fname):
    """Eventually remove .pyc and .pyo files associated to a Python script"""
    if osp.splitext(fname)[1] == '.py':
        for ending in ('c', 'o'):
            if osp.exists(fname + ending):
                os.remove(fname + ending)


def remove_pid():
    """Load navigator process ID."""
    check = True
    try:
        os.remove(PIDFILE)
    except Exception:  # pylint: disable=broad-except
        check = False
    return check


def set_windows_appusermodelid():
    """Make sure correct icon is used on Windows 7 taskbar"""
    try:
        from ctypes import windll  # pylint: disable=import-outside-toplevel
        name = 'anaconda.Anaconda-Navigator'
        return windll.shell32.SetCurrentProcessExplicitAppUserModelID(name)
    except AttributeError:
        return 'SetCurrentProcessExplicitAppUserModelID not found'



# -*- coding: utf-8 -*-

# pylint: disable=invalid-name,no-name-in-module,too-many-lines,unused-argument

# -----------------------------------------------------------------------------
# Copyright 2016 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
"""Wigets module."""

# yapf: disable

# Third party imports
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import QFrame, QLabel, QPushButton, QSizePolicy, QToolButton, QWidget


# yapf: enable


# --- Base widgets
# -----------------------------------------------------------------------------
class WidgetBase(QWidget):
    """Widget base implementation."""

    sig_hovered = Signal(bool)
    sig_focused = Signal(bool)

    def _fix_style(self):
        # Mac related issues
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)
        self.setFocusPolicy(Qt.StrongFocus)

    def _fix_layout(self, layout):
        if layout:
            layout.setSpacing(0)
            layout.setContentsMargins(0, 0, 0, 0)

            items = (layout.itemAt(i).widget() for i in range(layout.count()))
            for w in items:
                if w:
                    new_layout = w.layout()
                    self._fix_layout(new_layout)

    def focusInEvent(self, event):
        """Override Qt method."""
        QWidget.focusInEvent(self, event)
        self.setProperty('focused', True)

    def focusOutEvent(self, event):
        """Override Qt method."""
        QWidget.focusOutEvent(self, event)
        self.setProperty('focused', False)

    def enterEvent(self, event):
        """Override Qt method."""
        QWidget.enterEvent(self, event)
        self.setProperty('hovered', True)

    def leaveEvent(self, event):
        """Override Qt method."""
        QWidget.leaveEvent(self, event)
        self.setProperty('hovered', False)

    def setDisabled(self, value):
        """Override Qt method."""
        QWidget.setDisabled(self, value)
        self.setProperty('disabled', value)
        self.setProperty('enabled', not value)

    def setEnabled(self, value):
        """Override Qt method."""
        QWidget.setEnabled(self, value)
        self.setProperty('enabled', value)
        self.setProperty('disabled', not value)

    def setProperty(self, name, value):
        """Override Qt method."""
        QWidget.setProperty(self, name, value)
        self.style().unpolish(self)
        self.style().polish(self)
        self.update()

    def setLayout(self, layout):
        """Override Qt method."""
        self._fix_layout(layout)
        QWidget.setLayout(self, layout)


class ButtonBase(QPushButton, WidgetBase):
    """Base button used in CSS styling."""

    def __init__(self, *args, **kwargs):
        """Base button used in CSS styling."""
        super().__init__(*args, **kwargs)
        self.setAutoDefault(False)
        self.setDefault(False)
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)
        self.setFocusPolicy(Qt.StrongFocus)

    def mousePressEvent(self, event):
        """Override Qt method."""
        super().mousePressEvent(event)
        self.setProperty('pressed', True)

    def mouseReleaseEvent(self, event):
        """Override Qt method."""
        super().mouseReleaseEvent(event)
        self.setProperty('pressed', False)


class ButtonToolBase(QToolButton, WidgetBase):
    """Base button used in CSS styling."""

    def __init__(self, parent=None, text=''):
        """Base button used in CSS styling."""
        super().__init__(parent=parent)
        self.setCheckable(False)
        self.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
        self.setText(text)
        self.setFocusPolicy(Qt.StrongFocus)  # Needed on OSX
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)  # Needed on OSX
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)


class FrameBase(QFrame, WidgetBase):
    """Button used in CSS styling."""

    def __init__(self, *args, **kwargs):
        """Button used in CSS styling."""
        super().__init__(*args, **kwargs)
        self.setFocusPolicy(Qt.NoFocus)  # Needed on OSX
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)  # Needed on OSX

    def setLayout(self, layout):
        """Override Qt method."""
        self._fix_layout(layout)
        super().setLayout(layout)


class LabelBase(QLabel, WidgetBase):
    """Label used in CSS styling."""

    def __init__(self, *args, **kwargs):
        """Label used in CSS styling."""
        super().__init__(*args, **kwargs)
        self.setFocusPolicy(Qt.NoFocus)  # Needed on OSX
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)  # Needed on OSX


class ButtonNormal(ButtonBase):
    """Button used in CSS styling."""


class ButtonPrimary(ButtonBase):
    """Button used in CSS styling."""


class ButtonDanger(ButtonBase):
    """Button used in CSS styling."""


class SpacerHorizontal(LabelBase):
    """Label used in CSS styling."""


class SpacerVertical(LabelBase):
    """Label used in CSS styling."""



# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""Qt utilities."""

from __future__ import absolute_import, division, unicode_literals

import os

from qtpy.QtCore import QTimer  # pylint: disable=no-name-in-module
from qtpy.QtWidgets import QApplication  # pylint: disable=no-name-in-module

from navigator_updater.utils.linux_scaling import setup_scale_factor_for_linux


def qapplication(test_time=60):
    """Create QApplication instance."""
    app = QApplication.instance()

    if app is None:
        setup_scale_factor_for_linux()
        app = QApplication(['Anaconda-Navigator'])
        app.setApplicationName('Anaconda-Navigator')

    test_ci = os.environ.get('TEST_CI')

    if test_ci is not None:
        timer_shutdown = QTimer(app)
        timer_shutdown.timeout.connect(app.quit)
        timer_shutdown.start(test_time * 1000)
    return app
